from argparse import ArgumentParser
import json
import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from jsonschema.protocols import Validator

from cobra.asset.asset import get_asset
from cobra.utils.caches import TASK_CACHE
from cobra.utils.configurations import COBRA_VERSION
from cobra.utils.requests import session
from cobra.utils import schema
from cobra.utils.urls import TASK_URL
from cobra.utils.utils import find_key


def find_tasks(version: str = COBRA_VERSION,
               id: str = None,
               id_contains: str = None,
               ) -> Tuple[List[str], List[Dict[str, Any]]]:
    """
    Return a set of tasks UUIDs matching the given query parameters and a result dictionary that gives more details for

    each returned UUID.

    :param id: Task ID to query for
    :param id_contains: also match partial id such as 'simple/'
    :param version: Version of CoBRA to query in.
    :return: * List of UUIDs uniquely identifying Tasks in CoBRA
             * List of results giving more details incl. task_id, json, version, html visualization, solution count, and
               additional meta data.
    """
    query = [TASK_URL + "?"]
    if id is not None:
        query.append(f"scenario_id={id}")
    if id_contains is not None:
        query.append(f"scenario_id__icontains={id_contains}")
    if version is not None:
        query.append(f"version={version}")
    # tags: Set[str] = None,
    # if tags is not None:
    #     query.append(f"metadata__tags={tags}")
    # goal_types: Set[str] = None,
    query = query[0] + "&".join(query[1:])
    r = session.get(query)
    if r.status_code == 200:
        return [res["id"] for res in r.json()['results']], r.json()["results"]


def get_task_details(task_uuid: str) -> Dict[str, Any]:
    """
    Return the details of a task from the CoBRA API.

    :param task_uuid: The uuid of the task to return.
    :return: The details of the task as given by https://cobra.cps.cit.tum.de/api/ -> scenariocrok_read
    """
    r = session.get(TASK_URL + task_uuid + '/')
    if r.status_code != 200:
        raise ValueError(f"Task with uuid {task_uuid} not available.")
    return r.json()


def get_task(uuid: Optional[str] = None,
             id: Optional[str] = None,
             version: str = COBRA_VERSION,
             fetch_assets: bool = True,
             force_download: bool = False) -> Path:
    """Return a task with the given id and version."""
    if uuid is None and id is None:
        raise ValueError("Either uuid or id must be given.")

    if uuid is not None:
        if id is not None:
            logging.info("Both uuid and id were given. Ignoring id.")
        r_task_overview = session.get(TASK_URL + uuid + "/")
        if r_task_overview.status_code != 200:
            raise KeyError(f"Task with UUID = {uuid} not found.")
        id = r_task_overview.json()['scenario_id']
        task_overview = r_task_overview.json()
    base_path = TASK_CACHE.joinpath(version)
    task_file = base_path.joinpath(id + ".json")

    if not force_download and task_file.exists():
        if fetch_assets:
            for asset in find_key('file', json.load(task_file.open("r")))[0]:
                asset_location = base_path.joinpath(asset)
                if not asset_location.exists():
                    file = get_asset(asset, version=version)
                    asset_location.parent.mkdir(parents=True, exist_ok=True)
                    os.symlink(file, asset_location)
        return task_file

    if uuid is None:
        tasks_uuids, task_results = find_tasks(id=id, version=version)
        if len(tasks_uuids) == 0:
            raise KeyError(f"Task with id = {id} not found in version {version}.")
        elif len(tasks_uuids) > 1:
            raise KeyError(f"Multiple tasks with id = {id} found in version {version}.")
        uuid = tasks_uuids[0]
        task_overview = task_results[0]

    # Get json
    r_task_json = session.get(task_overview['json'])
    if r_task_json.status_code != 200:
        raise ValueError(f"Task json not available: {task_overview['json']}; Error code: {r_task_json.status_code}")
    task_file.parent.mkdir(parents=True, exist_ok=True)
    with open(task_file, 'wb') as f:
        f.write(r_task_json.content)

    # Get assets and symlink relative to task
    if fetch_assets:
        for asset in task_overview['assets']:
            asset_file_desired = base_path.joinpath(asset['name'])
            if asset_file_desired.exists():
                os.remove(asset_file_desired)
            asset_file_desired.parent.mkdir(parents=True, exist_ok=True)
            asset_file_desired.symlink_to(get_asset(asset['name'], version=asset['version'], force_download=False))
    return task_file


def upload_task(task: Path, user: str):
    """Upload a task to the CoBRA server."""
    raise NotImplementedError


def get_schema(version: str = COBRA_VERSION, force_download: bool = False) -> Path:
    """
    Return the schema file for tasks in the given version of the CoBRA benchmark.

    :param version: The version of the CoBRA benchmark to use.
    :param force_download: Whether to force download the schema file.
    :return: The schema file.
    :raise ValueError: If schema does not found for specific CoBRA version or other backend error encountered.
    """
    return schema.get_schema('TaskSchema', version, force_download)


def get_schema_validator(version: str = COBRA_VERSION,
                         force_download: bool = False) -> Tuple[Dict, Validator]:
    """
    Return the json schema for tasks in the given version of the CoBRA benchmark + a validator to check tasks with.

    :param version: The version of the CoBRA benchmark to use.
    :param force_download: Whether to force download the schema file.
    :return: The schema + jsonschema validator.
    :raise ValueError: If schema does not found for specific CoBRA version or other backend error encountered.
    """
    return schema.get_schema_validator('TaskSchema', version, force_download)


def main():
    """Main function to use task as commandline tool showing local copy of desired task ID."""
    parser = ArgumentParser(description='Download a cobra task via task ID and optional version.')
    parser.add_argument('taskID', help='ID of task', nargs=1, type=str)
    parser.add_argument('-v', '--version', help='CoBRA version to look for task in.', required=False,
                        default=COBRA_VERSION, type=str)
    parser.add_argument('-f', '--force_download', help='Force download of task and asset files.', action='store_true')
    args = parser.parse_args()
    print(get_task(id=args.taskID[0], version=args.version, force_download=args.force_download))


if __name__ == '__main__':
    main()



from getpass import getpass
from typing import Optional

from cobra.utils.urls import BASE_URL
from cobra.utils.requests import session


_refresh_token = None


def login(e_mail: str, password: Optional[str] = None) -> str:
    """
    Login to COBRA and return the access token.

    :param e_mail: The e-mail address of the user.
    :param password: The password of the user; alternatively, the user will be prompted to enter the password.
    :return: The access token.
    """
    global _refresh_token
    if _refresh_token is not None:
        r_refresh = session.post(BASE_URL + 'token/refresh/', data={'refresh': _refresh_token})
        if r_refresh.status_code == 200:
            return r_refresh.json()['access']
        # Fall through if refresh token is invalid.
    if password is None:
        password = getpass(f"Please enter the password for {e_mail}: ")
    r_login = session.post(BASE_URL + 'token/', data={'email': e_mail, 'password': password})
    if r_login.status_code != 200:
        raise ValueError("Invalid credentials.")
    _refresh_token = r_login.json()['refresh']
    return r_login.json()['access']


def logout():
    """
    Logout from COBRA.
    """
    global _refresh_token
    _refresh_token = None



import os
import unittest

import cobra.user
from cobra.utils.urls import BASE_URL
from cobra.utils.requests import session


class TestUserAPI(unittest.TestCase):
    """Test user API."""
    def test_user_login(self,
                        test_user=os.environ.get("TEST_USER", 'cobra@cps.cit.tum.de'),
                        test_password=os.environ.get("TEST_USER_PASSWORD", '!2345678')):
        """Test user.login API."""
        cobra.user.logout()  # Might have been logged in from previous test
        self.assertIsNone(cobra.user._refresh_token)
        with self.assertRaises(ValueError):
            _ = cobra.user.login('totally-made-up-user@not-an-email.com', 'totally-made-up-password')
        token = cobra.user.login(test_user, test_password)
        self.assertIsNotNone(token)
        r = session.post(BASE_URL + 'token/verify/', data={'token': token})
        self.assertEqual(r.status_code, 200)
        self.assertIsNotNone(cobra.user._refresh_token)
        token = cobra.user.login('m@tum.de')
        r = session.post(BASE_URL + 'token/verify/', data={'token': token})
        self.assertEqual(r.status_code, 200)



from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union


def find_key(key: str, data: Union[Dict, Sequence], modifier: Optional[Callable[[Any], Any]] = None,
             search_depth: int = 20) -> Tuple[List[Any], List[Any]]:
    """
    Find all appearances of key in a nested dict / sequence combination and return references to each.

    :param key: The key to search for.
    :param data: The data to search in.
    :param modifier: An optional modifier function to apply to each found data.
    :param search_depth: The maximum depth to search in.
    :return: A tuple of lists of all found values with matching key before and after the applied modifier.
    :note: Based on
      https://stackoverflow.com/questions/9807634/find-all-occurrences-of-a-key-in-nested-dictionaries-and-lists
    """
    if search_depth < 0:
        raise RecursionError("Maximum search depth exceeded.")
    matched_values = []
    post_modification_values = []
    for k, v in (
            data.items() if isinstance(data, Dict) else enumerate(data) if isinstance(data, Sequence) else []):
        if k == key:
            matched_values.append(data[k])
            if modifier is not None:
                data[k] = modifier(v)
            post_modification_values.append(data[k])
        elif isinstance(v, (Dict, Sequence)) and not isinstance(v, (str, bytes)):
            res = find_key(key, v, modifier, search_depth - 1)
            matched_values += res[0]
            post_modification_values += res[1]
    return matched_values, post_modification_values


def map2path(p: Union[str, Path]) -> Path:
    """Maps strings to paths, but does nothing else s.t., e.g., Django paths are not cast."""
    if isinstance(p, str):
        return Path(p)
    return p



from pathlib import Path
import shutil

from cobra.utils.configurations import COBRA_CONFIG


MAIN_CACHE_DIR = Path(COBRA_CONFIG['FILE_LOCATIONS']['cache'])
ASSET_CACHE = MAIN_CACHE_DIR.joinpath('assets')
MODULE_DB_CACHE = MAIN_CACHE_DIR.joinpath('module_dbs')
TASK_CACHE = MAIN_CACHE_DIR.joinpath('tasks')
SOLUTION_CACHE = MAIN_CACHE_DIR.joinpath('solutions')
SCHEMA_CACHE = MAIN_CACHE_DIR.joinpath('schemas')
caches = {MAIN_CACHE_DIR, ASSET_CACHE, MODULE_DB_CACHE, TASK_CACHE, SOLUTION_CACHE, SCHEMA_CACHE}


def generate_caches():
    """Creates the cache directories."""
    for dir in caches:
        dir.mkdir(parents=True, exist_ok=True)


def reset_cache():
    """Deletes the cache directory."""
    from cobra.utils.requests import session
    for dir in caches:
        if dir is MAIN_CACHE_DIR:
            continue
        shutil.rmtree(dir, ignore_errors=True)
    generate_caches()
    session.cache.clear()


generate_caches()


if __name__ == '__main__':
    val = input("This will delete the cache directory. Press y and enter to continue.")
    if val == 'y':
        reset_cache()
    else:
        print("Aborted.")



import logging
import time
import unittest

from cobra.utils.urls import BASE_URL
from cobra.utils.requests import session


class TestUtils(unittest.TestCase):
    """Test utils."""
    def test_API_available(self):
        """Test if API is available."""
        logging.info(f"Testing if API is available at {BASE_URL}.")
        r = session.get(BASE_URL)
        self.assertEqual(r.status_code, 200)

    def test_rate_limit(self, per_second=5):
        """Check that limited to 5 per second (default)."""
        with session.cache_disabled():  # Otherwise cache will hide rate limiting
            time.sleep(10)  # Wait for local and server rate limit to reset
            t0 = time.time()
            # Make sure over burst
            print("Saturate")
            for i in range(per_second * 2):
                r = session.get(BASE_URL)
                print(f"{time.time() - t0:.2f}")
            # Measure time during limiting
            print("Limiting")
            t2 = time.time()
            for i in range(per_second * 2):
                r = session.get(BASE_URL)
                self.assertEqual(r.status_code, 200)
                print(f"{time.time() - t0:.2f}")
            t3 = time.time()
            self.assertGreater(t3 - t2, 1.5)  # Need significantly more than a single cycle



from cobra.utils.caches import ASSET_CACHE
from cobra.utils.configurations import COBRA_VERSION
from cobra.utils.urls import ASSET_URL
from cobra.utils.requests import session


def get_asset(name, version=COBRA_VERSION, force_download=False):
    """
    Returns asset from CoBRA if it exists.

    :param name: Name of asset to look for.
    :param version: Version of CoBRA to find this asset for.
    :param force_download: If True, asset will be downloaded even if it exists in cache.
    :return: Path to asset file.
    :raises FileNotFoundError: If asset does not exist in CoBRA of specific version.
    """
    asset_file = ASSET_CACHE.joinpath(version).joinpath(name)
    if not force_download and asset_file.exists():
        return asset_file
    asset_URI = f"{ASSET_URL}{name}?version={version}"
    r = session.get(asset_URI)
    if r.status_code == 200:
        asset_file.parent.mkdir(parents=True, exist_ok=True)
        with open(asset_file, 'wb') as f:
            f.write(r.content)
        return asset_file
    else:
        raise FileNotFoundError(f"Asset {name} (version {version}) does not exist in CoBRA.")



import json
import unittest
import xml.etree.ElementTree as ET

import pinocchio
import timor.Robot

import cobra
from cobra.utils import urls, logging, caches
from cobra.utils.requests import session


class TestRobotAPI(unittest.TestCase):
    """Test robot API."""
    def test_API_available(self):
        """Test if API is available."""
        r = session.get(urls.MODULE_DB_URL)
        logging.info(f"Testing if robot API is available at {r.request.url}.")
        self.assertEqual(r.status_code, 200)
        self.assertGreaterEqual(len(r.json()), 1)

    def test_all_downloadable(self):
        """Test if all robots are downloadable."""
        caches.reset_cache()
        robots = cobra.robot.get_available_module_dbs()
        for robot in robots:
            f = cobra.robot.get_module_db(robot)
            self.assertTrue(f.is_file())
            data = json.load(f.open())
            self.assertTrue('modules' in data)
            self.assertTrue(len(data['modules']) > 0)

        local_robots = cobra.robot.get_available_module_dbs(locally=True)
        self.assertEqual(robots, local_robots)
        for robot in robots:
            f = cobra.robot.get_module_db(robot)
            self.assertTrue(f.is_file())
            data = json.load(f.open())
            self.assertTrue('modules' in data)
            self.assertTrue(len(data['modules']) > 0)

    def test_schema_download(self):
        caches.reset_cache()
        f = cobra.robot.get_schema()
        self.assertTrue(f.is_file())
        data = json.load(f.open())
        self.assertTrue('$id' in data)
        self.assertTrue('$schema' in data)

    def test_schema_validator(self):
        schema, validator = cobra.robot.get_schema_validator()
        self.assertIsNotNone(schema)
        self.assertIsNotNone(validator)
        db_file = cobra.robot.get_module_db('IMPROV')
        db_data = json.load(db_file.open())
        self.assertTrue(validator.is_valid(db_data))

    def test_urdf_download(self):
        caches.reset_cache()
        f = cobra.robot.get_urdf('IMPROV', module_order=['1', '21', '101'])
        self.assertTrue(f.is_file())
        data = f.open().read()
        self.assertTrue(data.startswith('<robot name="Auto-Generated: '))
        f1 = cobra.robot.get_urdf('IMPROV', module_order=['1', '21', '101'])  # Hit cache
        self.assertTrue(f1.is_file())
        self.assertEqual(f, f1)
        # Make sure assets available
        urdf_tree = ET.parse(f)
        for mesh in urdf_tree.findall('.//mesh'):
            file_name = mesh.attrib['filename']
            file_name = file_name.replace('package://', '')
            self.assertTrue(f.parent.joinpath(file_name).is_file())
        # Make sure URDF is readable
        r_timor = timor.Robot.PinRobot.from_urdf(f, f.parent)
        self.assertIsNotNone(r_timor.random_configuration())
        r_pin = pinocchio.RobotWrapper.BuildFromURDF(str(f), str(f.parent))
        self.assertIsNotNone(r_pin)

        urdf_with_wrl = cobra.robot.get_urdf('IMPROV', module_order=['1', '21', '101'],
                                             force_download=True, accept_wrl=True)
        self.assertTrue(urdf_with_wrl.is_file())
        for mesh in ET.parse(urdf_with_wrl).findall('.//mesh'):
            file_name = mesh.attrib['filename']
            file_name = file_name.replace('package://', '')
            self.assertTrue(f.parent.joinpath(file_name).is_file())
        with self.assertRaises(ValueError):
            timor.Robot.PinRobot.from_urdf(urdf_with_wrl, f.parent)  # Timor / pinochio does not support wrl



from cobra.utils import configurations

BASE_URL = configurations.COBRA_CONFIG['API']['base_url']
ASSET_URL = BASE_URL + 'scenariocrok/assets/'
MODULE_DB_URL = BASE_URL + 'robots'
TASK_URL = BASE_URL + 'scenariocrok/'
SOLUTION_URL = BASE_URL + 'solutioncrok/'
RANKING_URL = SOLUTION_URL + 'ranking/'
SCHEMA_URL = BASE_URL + 'schemas/'



import os
from pathlib import Path
import re
import sys

import nbconvert
import nbformat
import pytest

from cobra.user.user import login


sys.path.append(str(Path(__file__).parent.parent.joinpath('tutorials')))  # for local imports in tutorials

excluded_notebooks = {  # Notebooks that are not runnable automatically
    'Tolerated_Poses.ipynb'
}
notebooks_with_upload = {  # Notebooks that need login information
    '0_CoBRA_Introduction.ipynb',
    '1_CoBRA_Timor_Interaction.ipynb',
}


normal_notebooks = tuple(p for p in Path(__file__).parent.parent.glob('tutorials/*.ipynb')
                         if p.name not in excluded_notebooks | notebooks_with_upload)
notebooks_with_upload = tuple(p for p in Path(__file__).parent.parent.glob('tutorials/*.ipynb')
                              if p.name not in excluded_notebooks and p.name in notebooks_with_upload)
nb2script = nbconvert.ScriptExporter()


def _generate_script(nb):
    notebook = nbformat.read(nb.open("r"), as_version=4)
    script, _ = nb2script.from_notebook_node(notebook)
    script = re.sub(r'get_ipython\(\).*$\n', '', script, flags=re.MULTILINE)  # Remove ! commands
    script = re.sub(r'^.*condacolab.*$\n', '', script, flags=re.MULTILINE)  # Remove conda magic
    script = re.sub(r'^.*delete_solution.*$\n', '', script, flags=re.MULTILINE)  # Remove delete calls
    return script


def _exec_script(script):
    """Execute script in a clean context."""
    exec(script)


@pytest.mark.parametrize("nb", normal_notebooks)
def test_all_runable(nb: Path):
    # Make notebook into script
    script = _generate_script(nb)
    # Run script
    _exec_script(script)  # To keep local context clean


@pytest.mark.parametrize("nb", notebooks_with_upload)
def test_notebook_with_upload(
        nb: Path,
        test_user=os.environ.get("TEST_USER", 'cobra@cps.cit.tum.de'),
        test_password=os.environ.get("TEST_USER_PASSWORD", '!2345678')):
    """Test if notebook with upload works."""
    script = _generate_script(nb)
    script = script.replace('your-name', 'nb_test')
    script = script.replace('your-email', test_user)
    script = script.replace('your-affiliation', 'unit_test')
    login(test_user, test_password)  # Make sure user token already set
    _exec_script(script)



from contextlib import redirect_stdout
import io
import json
import logging
from pathlib import Path
import sys
from unittest.mock import patch

from timor.task import Task
import unittest

import cobra.task
from cobra.utils import caches, urls
from cobra.utils.configurations import COBRA_VERSION
from cobra.utils.requests import session


class TestTaskAPI(unittest.TestCase):
    """Test task API."""
    def test_API_available(self):
        """Test if API is available."""
        r = session.get(urls.TASK_URL + '?page=1&page_size=2')
        logging.info(f"Testing if API is available at {r.request.url}.")
        self.assertEqual(r.status_code, 200)
        self.assertGreaterEqual(len(r.json()['results']), 1)

        r = session.get(urls.TASK_URL + '?scenario_id=simple/PTP_1&version=2022')
        logging.info(f"Testing if API is available at {r.request.url}.")
        self.assertEqual(r.status_code, 200)
        self.assertEqual(len(r.json()['results']), 1)

    def test_task_retrieval(self):
        caches.reset_cache()
        task_file = cobra.task.get_task(id='simple/PTP_1')
        self.assertTrue(task_file.is_file())
        self.assertIsNotNone(json.load(task_file.open()))
        task = Task.Task.from_json_file(task_file)
        self.assertIsNotNone(task)
        with self.assertRaises(ValueError):
            cobra.task.get_task(version='some-not-existing-version')

    def test_find_tasks(self):
        caches.reset_cache()
        candidates, details = cobra.task.find_tasks(id='simple/PTP_1')
        self.assertEqual(len(candidates), 1)
        # Check details
        self.assertEqual(len(details), 1)
        self.assertEqual(details[0]['scenario_id'], 'simple/PTP_1')
        self.assertEqual(details[0]['version'], COBRA_VERSION)
        self.assertEqual(session.get(details[0]['html']).status_code, 200)
        self.assertEqual(session.get(details[0]['json']).status_code, 200)
        self.assertEqual(details[0]['metadata']['goal_count'], 2)
        task_file = cobra.task.get_task(uuid=details[0]['id'])
        self.assertTrue(task_file.is_file())
        self.assertIsNotNone(json.load(task_file.open()))
        task = Task.Task.from_json_file(task_file)
        self.assertIsNotNone(task)
        # Check details same:
        detail = cobra.task.get_task_details(candidates[0])
        for k in detail.keys():
            if k in {'json', 'png', 'html'}:
                continue  # Might change due to storage bucket creating new access tokens
            self.assertEqual(detail[k], details[0][k])
        # Check wrong version
        candidates, details = cobra.task.find_tasks(id='simple/PTP_1', version='some-not-existing-version')
        self.assertEqual(len(candidates), 0)
        # Check wrong id
        candidates, details = cobra.task.find_tasks(id='some-not-existing-id')
        self.assertEqual(len(candidates), 0)

    def test_task_with_asset(self):
        caches.reset_cache()
        task_file = cobra.task.get_task(id='simple/PTP_2', fetch_assets=False)
        with self.assertRaises(FileNotFoundError):
            Task.Task.from_json_file(task_file)
        task_file = cobra.task.get_task(id='simple/PTP_2', fetch_assets=True)
        task = Task.Task.from_json_file(task_file)
        self.assertIsNotNone(task)

    def test_non_exising_task(self):
        with self.assertRaises(KeyError):
            cobra.task.get_task(uuid='some-not-existing-uuid')
        with self.assertRaises(ValueError):
            cobra.task.get_task_details(task_uuid='some-not-existing-uuid')
        with self.assertRaises(KeyError):
            cobra.task.get_task(id='some-not-existing-id')

    def test_schema_download(self):
        caches.reset_cache()
        f = cobra.task.get_schema()
        self.assertTrue(f.is_file())
        data = json.load(f.open())
        self.assertTrue('$id' in data)
        self.assertTrue('$schema' in data)
        # Check that cached version is used
        f = cobra.task.get_schema()
        self.assertTrue(f.is_file())
        data = json.load(f.open())
        self.assertTrue('$id' in data)
        self.assertTrue('$schema' in data)

    def test_schema_validator(self):
        schema, validator = cobra.task.get_schema_validator()
        self.assertIsNotNone(schema)
        self.assertIsNotNone(validator)
        task_file = cobra.task.get_task(id='simple/PTP_2', fetch_assets=False)
        task_data = json.load(task_file.open())
        self.assertTrue(validator.is_valid(task_data))

    def test_main(self):
        caches.reset_cache()
        stdout = io.StringIO()
        with redirect_stdout(stdout):
            with patch.object(sys, 'argv', ["solution.py", "simple/PTP_1"]):
                cobra.task.main()
        task_file = Path(stdout.getvalue().strip())
        self.assertTrue(task_file.is_file())
        task = Task.Task.from_json_file(task_file)
        self.assertIsNotNone(task)



"""No external dependencies here - the imports themselves are to be tested"""
import importlib
from pathlib import Path
import re
import sys
import unittest

import pytest

from cobra.utils.urls import BASE_URL
from cobra.utils.requests import session


@pytest.mark.imports
class TestImportingEverything(unittest.TestCase):
    """Tests that every module can be imported properly without any cyclic dependencies, etc."""

    exclude = ()  # Regular expressions for files to be excluded from the unittest

    def setUp(self) -> None:
        self.file = Path(__file__)
        self.tests = self.file.parent
        self.src_dir = self.tests.parent
        self.py_files = list(path.relative_to(self.src_dir) for path in self.src_dir.rglob('*.py'))
        imports = tuple(re.sub(r'/', '.', re.sub(r'\.py', '', str(file))) for file in self.py_files)
        self.imports = tuple((import_ for import_ in imports if not any(re.match(exclude, import_)
                                                                        for exclude in self.exclude)))

    def test_imports(self):
        _original_modules = tuple(sys.modules.keys())
        ignore = ('pinocchio',)  # External libraries with problems when deleted and reloaded repeatedly
        for import_as in self.imports:
            if ('__init__' in import_as) or (import_as.split('.')[0] in ('tests', 'demos', 'doc',
                                                                         'build', 'tutorials')):
                # Those will never be imported
                continue
            if import_as.startswith('src'):  # import from cobra, not from src folder
                import_as = import_as[4:]

            importlib.import_module(import_as)  # actual test

            # undo the import of local packages to be independent of previous imports in this test
            for key in tuple(sys.modules.keys()):
                if key not in _original_modules and not any(lib in key for lib in ignore):
                    del sys.modules[key]

    def test_default_API(self):
        """Test that API available if installed from scratch."""
        r = session.get(BASE_URL)
        self.assertEqual(r.status_code, 200)


if __name__ == '__main__':
    unittest.main()



from requests import Session
from requests.adapters import HTTPAdapter
from requests_ratelimiter import LimiterMixin
from requests_cache import CacheMixin
from urllib3 import Retry

from cobra.utils.caches import MAIN_CACHE_DIR


class CachedLimiterSession(CacheMixin, LimiterMixin, Session):
    """Session class with caching and rate-limiting behavior.

    Accepts keyword arguments for both LimiterSession and CachedSession.
    """


"""
Main session object to use when interacting with any part of the CoBRA API.

Features:
 - Rate limiting: <= 5 per second and <= 100 per minute
"""
session = CachedLimiterSession(str(MAIN_CACHE_DIR.joinpath('requests_cache.db')),
                               per_second=5, per_minute=100,
                               expire_after=360, allowable_methods=('GET', 'HEAD'),
                               allowable_codes=(200, ))

retries = Retry(  # Automatically retry on rate limiting
    total=3,
    backoff_factor=.1,
    status_forcelist=[429, ]
)


session.mount('https://', HTTPAdapter(max_retries=retries))



import argparse
import io
import json
import re
from pathlib import Path
import shutil
from typing import Optional, Tuple, Union, Dict, List, Any
import zipfile

from jsonschema.protocols import Validator

from cobra.user.user import login
from cobra.utils.caches import SOLUTION_CACHE
from cobra.utils.configurations import COBRA_VERSION
import cobra.utils.logging as logging
from cobra.utils.requests import session
from cobra.task.task import find_tasks, get_task
from cobra.utils import schema
from cobra.utils.urls import RANKING_URL, SOLUTION_URL
from cobra.utils.utils import map2path


def find_solutions(task_id: Optional[str] = None, task_uuid: Optional[str] = None, version: str = COBRA_VERSION,
                   cost_function: Optional[str] = None) -> Tuple[List[str], List[Dict[str, Any]]]:
    """
    Return a list of solutions for the given task.

    :param task_id: The id of the task the solutions belong to.
    :param task_uuid: The uuid of the task the solutions belong to.
    :param version: The version of the CoBRA benchmark to use.
    :param cost_function: The cost function to filter for.
    :return: A list of solution uuids + their details from the database.
    """
    if task_uuid is None and task_id is None:
        raise ValueError("Either task_uuid or task_id must be given.")
    if task_uuid is None:
        _, details = find_tasks(id=task_id, version=version)
        if len(details) != 1:
            raise KeyError(f"Task with id {task_id} not unique in version {version}.")
        task_uuid = details[0]['id']
    query = RANKING_URL + task_uuid + "/"
    if cost_function is not None:
        query += f"?costFunction__icontains={cost_function}"
    r = session.get(query)
    if r.status_code != 200:
        raise ValueError(f"Solutions for task with uuid {task_uuid} not found.")
    if r.json()['next'] is not None:
        logging.warning(f"Only the first {r.json()['count']} solutions are returned.")
    return [res["id"] for res in r.json()['results']], r.json()['results']


def get_solution_details(solution_uuid: str) -> Dict[str, Any]:
    """
    Return the details of a solution from the CoBRA API.

    :param solution_uuid: The uuid of the solution to return.
    :return: The details of the solution as given by https://cobra.cps.cit.tum.de/api/ -> solutioncrok_read
    """
    r = session.get(SOLUTION_URL + solution_uuid + '/')
    if r.status_code != 200:
        raise ValueError(f"Solution with uuid {solution_uuid} not available.")
    return r.json()


def get_solution(solution_uuid: str, force_download: bool = False) -> Tuple[Path, Path]:
    """
    Return a solution file with the given uuid.

    :param solution_uuid: The uuid of the solution to return.
    :param force_download: Whether to force download the solution file.
    :return: The solution file and associated task file.
    """
    solution_file = SOLUTION_CACHE.joinpath(solution_uuid + ".json")
    if solution_file.exists() and not force_download:
        logging.debug("Solution file already exists. Returning cached version.")
        solution_data = json.load(solution_file.open("r"))
        return solution_file, get_task(id=solution_data["taskID"], version=solution_data["version"])

    r = session.get(SOLUTION_URL + solution_uuid + "/")
    if r.status_code != 200:
        raise ValueError(f"Solution with uuid {solution_uuid} not available.")
    r_json = session.get(r.json()["json"])
    if r_json.status_code != 200:
        raise ValueError(f"Solution json for uuid {solution_uuid} not available.")

    with open(solution_file, "wb") as f:
        f.write(r_json.content)

    task_file = get_task(uuid=r.json()["scenario_id"], force_download=force_download)

    return solution_file, task_file


def submit_solution(solution_file: Union[Path, str], user_email: str,
                    password: Optional[str] = None) -> Tuple[bool, str]:
    """
    Submit a solution file to the CoBRA benchmark.

    :param solution_file: The solution file to submit.
    :param user_email: The email of the user submitting the solution.
    :param password: The password of the user submitting the solution. Alternative enter via CLI input.
    :return: True if successful + UUID of solution.
    """
    access_token = login(user_email, password)
    r = session.post(SOLUTION_URL + 'solutionsupload/upload/',
                     files={"json": map2path(solution_file).open("rb")},
                     headers={"Authorization": f"Bearer {access_token}"})
    if r.status_code != 201:
        if r.status_code == 400 and b'Set overwrite=True in order to overwrite' in r.content:
            logging.info("Solution already exists. Did not overwrite.")
            uuid = re.findall("((?<=uuid: )[0-9a-f-]+)", str(r.content))
            if len(uuid) == 1:
                return True, uuid[0]
        raise ValueError(f"Solution upload failed with status code {r.status_code}; details: {r.content}.")
    return True, r.json()['id']


def delete_solution(solution_uuid: str, user_email: str, password: Optional[str] = None) -> bool:
    """
    Delete solution with given uuid.

    :param solution_uuid: The uuid of the solution to delete.
    :param user_email: The email of the user deleting the solution.
    :param password: The password of the user deleting the solution. Alternative enter via CLI input.
    :return: True if successful.
    :raises ValueError: If deletion fails.
    """
    access_token = login(user_email, password)
    r = session.delete(SOLUTION_URL + f"{solution_uuid}/",
                       headers={"Authorization": f"Bearer {access_token}"})
    if r.status_code != 204:
        raise ValueError(f"Solution deletion failed with status code {r.status_code}; details: {r.content}.")
    session.cache.delete(urls=[SOLUTION_URL + f"{solution_uuid}/", ])
    solution_file = SOLUTION_CACHE.joinpath(solution_uuid + ".json")
    solution_file.unlink(missing_ok=True)  # Also remove from cache
    return True


def get_schema(version: str = COBRA_VERSION, force_download: bool = False) -> Path:
    """
    Return the schema file for solutions in the given version of the CoBRA benchmark.

    :param version: The version of the CoBRA benchmark to use.
    :param force_download: Whether to force download the schema file.
    :return: The schema file.
    :raise ValueError: If schema does not found for specific CoBRA version or other backend error encountered.
    """
    return schema.get_schema('SolutionSchema', version, force_download)


def get_schema_validator(version: str = COBRA_VERSION,
                         force_download: bool = False) -> Tuple[Dict, Validator]:
    """
    Return the json schema for CoBRA solutions in the given version of the CoBRA benchmark + a validator to check with.

    :param version: The version of the CoBRA benchmark to use.
    :param force_download: Whether to force download the schema file.
    :return: The schema + jsonschema validator.
    :raise ValueError: If schema does not found for specific CoBRA version or other backend error encountered.
    """
    return schema.get_schema_validator('SolutionSchema', version, force_download)


def get_negative_solutions(force_download: bool = False) -> Tuple[Path]:
    """
    Get negative solution files from CoBRA.

    These include an extra human-readable field `failure` describing why this specific solution is considered invalid.

    :return: Folder containing negative solutions.
    """
    goal_path = SOLUTION_CACHE.joinpath("negativesolutions")
    if not force_download and goal_path.exists():
        logging.debug("Negative solutions already downloaded. Returning cached version.")
        return tuple(goal_path.rglob("**/*.json"))

    if goal_path.exists():
        shutil.rmtree(goal_path)

    r = session.get("https://cobra.cps.cit.tum.de/api/negativesolutions")
    if r.status_code != 200:
        raise ValueError(f"Negative solutions not available - {r.status_code}; details: {r.content}.")
    zipfile.ZipFile(io.BytesIO(r.content)).extractall(goal_path)
    return tuple(goal_path.rglob("**/*.json"))


def main():
    """Entry point to use solution as a command line tool for solution overview and file retrieval."""
    parser = argparse.ArgumentParser(description="Download solutions from the CoBRA benchmark.")
    parser.add_argument("--task_id", type=str, help="The id of the task to download solutions for.", default=None)
    parser.add_argument("--task_uuid", type=str, help="The uuid of the task to download solutions for.", default=None)
    parser.add_argument("--solution_uuid", type=str, help="The uuid of the solution to download.", default=None)
    parser.add_argument("--version", type=str, default=COBRA_VERSION, help="The version of the CoBRA benchmark to use.")
    parser.add_argument("--cost_function", type=str, default=None,
                        help="The cost function to use when querying solutions to a task.")
    args = parser.parse_args()

    if args.task_id is None and args.task_uuid is None and args.solution_uuid is None:
        raise ValueError("Either task_id, task_uuid, or solution_uuid must be given.")

    if args.solution_uuid is not None:
        if args.task_id is not None or args.task_uuid is not None:
            logging.warning("Ignoring task_id and task_uuid as solution_uuid is given.")
        solution, task = get_solution(args.solution_uuid)
        print(f"Solution file: {solution}, Task file: {task}")
    else:
        solution_uuids, solutions = find_solutions(task_id=args.task_id, task_uuid=args.task_uuid,
                                                   version=args.version, cost_function=args.cost_function)
        print(f"Found {len(solutions)} solutions for task {args.task_id} in version {args.version}.")
        print("Solution id                          | Cost     | Cost function")
        print("-------------------------------------+----------+--------------")
        for solution in solutions:
            print(f"{solution['id']} | {solution['cost']:8.2f} | {solution['costFunction']}")
    return 0


if __name__ == "__main__":
    main()



import unittest

from timor.Geometry import Mesh

from cobra.asset.asset import get_asset
from cobra.utils import caches


class TestAssetsAPI(unittest.TestCase):
    """Test robot API."""
    def test_asset_download(self):
        caches.reset_cache()
        file = get_asset('assets/DMU_125P.stl')
        mesh = Mesh({'file': file})
        self.assertIsNotNone(mesh)
        self.assertGreater(mesh.measure_volume, 0)

    def test_asset_fail(self):
        with self.assertRaises(FileNotFoundError):
            get_asset('some-really-non-existent-file')



import json
from pathlib import Path
from typing import Dict, Tuple

from jsonschema.protocols import Validator
import jsonschema.validators
import referencing
import referencing.exceptions

from cobra.utils.caches import SCHEMA_CACHE
from cobra.utils.configurations import COBRA_VERSION
from cobra.utils.requests import session
from cobra.utils.urls import SCHEMA_URL


def retrieve_schema_uri(schema_uri: str) -> referencing.Resource:
    """Retrieve a schema from a URI; cached via session."""
    response = session.get(schema_uri)
    if response.status_code != 200:
        raise referencing.exceptions.NoSuchResource(schema_uri)
    return referencing.Resource.from_contents(response.json())


"""
Registry for all referenced schemas using global session to manage caching and retries.
"""


SCHEMA_REGISTRY = referencing.Registry(retrieve=retrieve_schema_uri)


def get_schema(schema: str, version: str = COBRA_VERSION, force_download: bool = False) -> Path:
    """
    Download a schema from the CoBRA API

    :param schema: The schema to download.
    :param version: The version of CoBRA to download from.
    :param force_download: Whether to force a download even if the schema is already cached.
    :return: The path to the downloaded schema.
    """
    if version != COBRA_VERSION:
        raise NotImplementedError
    schema_file = SCHEMA_CACHE.joinpath(f"{schema}.json")
    if not force_download and schema_file.exists():
        return schema_file
    r = session.get(SCHEMA_URL + f"{schema}.json")
    if r.status_code != 200:
        raise ValueError(f"Could not get schema {schema} from CoBRA (error {r.status_code}).")
    schema_file.write_text(r.text)
    return schema_file


def get_schema_validator(schema: str, version: str = COBRA_VERSION,
                         force_download: bool = False) -> Tuple[Dict, Validator]:
    """Return a loaded schema and a validator for it. Retrieval is done via get_schema."""
    schema_file = get_schema(schema, version, force_download)
    with open(schema_file) as f:
        main_schema = json.load(f)

    # Allow also tuple as json array
    new_type_checker = jsonschema.Draft202012Validator.TYPE_CHECKER.redefine(
        "array", lambda _, instance:
        jsonschema.Draft202012Validator.TYPE_CHECKER.is_type(instance, "array") or isinstance(instance, tuple))
    validator_with_tuple = jsonschema.validators.extend(jsonschema.Draft202012Validator, type_checker=new_type_checker)
    validator = validator_with_tuple(main_schema, registry=SCHEMA_REGISTRY)

    return main_schema, validator



import io
from pathlib import Path
from typing import List, Set, Tuple, Dict

from jsonschema.protocols import Validator
import zipfile

from cobra.utils.caches import MODULE_DB_CACHE
from cobra.utils.configurations import COBRA_VERSION
from cobra.utils import logging
from cobra.utils.requests import session
from cobra.utils import schema
from cobra.utils.urls import MODULE_DB_URL


def get_module_db(name: str, version: str = COBRA_VERSION, force_download: bool = False) -> Path:
    """
    Returns module database from CoBRA if it exists.

    :param name: Name of module database to look for.
    :param version: Version of CoBRA to find this module database for.
    :param force_download: If True, module database will be downloaded even if it exists in cache.
    :return: Path to module database file.
    :raises FileNotFoundError: If module database does not exist in CoBRA of specific version.
    """
    module_db_dir = MODULE_DB_CACHE.joinpath(version).joinpath(name)
    module_db_file = module_db_dir.joinpath('modules.json')
    if not force_download and module_db_file.exists():
        return module_db_file
    module_db_uri = f"{MODULE_DB_URL}/{name}?zip=True"
    logging.debug(f"Trying to get {module_db_uri}")
    r = session.get(module_db_uri)
    logging.debug(f"Request status: {r.status_code}")
    if r.status_code == 200:
        module_db_dir.mkdir(parents=True, exist_ok=True)
        zipfile.ZipFile(io.BytesIO(r.content)).extractall(module_db_dir)
        return module_db_file
    else:
        raise FileNotFoundError(f"Asset {name} (version {version}) does not exist in CoBRA.")


def get_available_module_dbs(version: str = COBRA_VERSION, locally: bool = False) -> Set[str]:
    """
    Returns a set of module databases available in CoBRA of specific version.

    :param version: Version of CoBRA to find module databases for.
    :param locally: If True, will return module databases available in cache.
    :return: Set of module databases available in CoBRA of specific version.
    :raises ValueError: If module databases could not be retrieved from CoBRA.
    """
    if locally:
        return {x.name for x in MODULE_DB_CACHE.joinpath(version).iterdir()}
    else:
        r = session.get(MODULE_DB_URL)
        if r.status_code == 200:
            return set(r.json())
        else:
            raise ValueError(f"Could not get available module databases from CoBRA (error {r.status_code}).")


def get_schema(version: str = COBRA_VERSION, force_download: bool = False) -> Path:
    """
    Get schema of the CoBRA module database.

    :param version: Version of CoBRA to find schema for.
    :param force_download: If True, schema will be downloaded even if it exists in cache.
    :return: Path to schema file.
    :raise ValueError: If schema does not found for specific CoBRA version or other backend error encountered.
    """
    return schema.get_schema('ModuleSchema', version, force_download)


def get_schema_validator(version: str = COBRA_VERSION,
                         force_download: bool = False) -> Tuple[Dict, Validator]:
    """
    Return the json schema for module DBs in the given version of the CoBRA benchmark + a validator to check DBs with.

    :param version: The version of the CoBRA benchmark to use.
    :param force_download: Whether to force download the schema file.
    :return: The schema + jsonschema validator.
    :raise ValueError: If schema does not found for specific CoBRA version or other backend error encountered.
    """
    return schema.get_schema_validator('ModuleSchema', version, force_download)


def get_urdf(module_db: str, module_order: List[str], version: str = COBRA_VERSION,
             force_download: bool = False, accept_wrl: bool = False) -> Path:
    """
    Get URDF of robot with specific module order. The assets are stored alongside the URDF.

    :param module_db: Name of module database to use.
    :param module_order: List of modules in order from base to end-effector.
    :param version: Version of CoBRA to find module database for.
    :param force_download: If True, URDF will be downloaded even if it exists in cache.
    :param accept_wrl: If True, will accept WRL files as well as stl files for URDF geometries.
    :return: Path to URDF file.
    :raise ValueError: If module database does not exist in CoBRA of specific version or other backend error.
    :note: For individual use only. If you want to create many URDFs please consider using timor-python locally.
    """
    get_module_db(module_db, version, force_download)  # Get module db such that assets are downloaded
    urdf_file = MODULE_DB_CACHE.joinpath(version).joinpath(module_db).joinpath(f"robot-{'-'.join(module_order)}.urdf")
    if urdf_file.exists() and not force_download:
        return urdf_file
    module_order_string = '"' + '", "'.join(module_order) + '"'
    query = f'{MODULE_DB_URL}/{module_db}?urdf=True&moduleOrder=[{module_order_string}]&wrl={accept_wrl}'
    r = session.get(query)
    if r.status_code != 200:
        raise ValueError(f"Could not get URDF from CoBRA (error {r.status_code}).")
    urdf_file.write_text(r.text)
    return urdf_file


def upload_task(task: Path, user: str):
    """Upload a task to the CoBRA server."""
    raise NotImplementedError



import logging as root_logging
import os
from pathlib import Path
from typing import Union

from cobra.utils.configurations import COBRA_CONFIG

"""
Drop in replacement for the basic logging functionality provided by logging.

Replaces the root logger by a custom configured logger.
"""


CRITICAL = root_logging.CRITICAL
ERROR = root_logging.ERROR
WARNING = root_logging.WARNING
INFO = root_logging.INFO
DEBUG = root_logging.DEBUG
NOTSET = root_logging.NOTSET


def getLogger() -> root_logging.Logger:
    """Get the custom logger."""
    return root_logging.getLogger('CoBRA I/O')


def setLevel(level):
    """Set the logger level."""
    getLogger().setLevel(level)


def basicConfig(filename: str = None,
                filemode: str = 'w',
                output_format: str = '%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                datefmt: str = '%Y-%m-%d %H:%M:%S',
                level: Union[int, str] = INFO):
    """
    Set logging configurations for the runtime environment

    :param filename: Name to store logs to
    :param filemode: Mode (append ('a'), overwrite ('w'), ...)
    :param output_format: Format string for log
    :param datefmt: Specify time in logs
    :param level: Logging level for the logger (use logging.DEBUG, WARN, ...)
    """
    if isinstance(level, str):
        try:
            level = int(level)
        except ValueError:
            level = root_logging.getLevelName(level)
    if filename is not None:
        try:
            Path(filename).parent.mkdir(exist_ok=True)
            handler = root_logging.FileHandler(filename=filename, mode=filemode)
        except PermissionError as e:
            print(f"There is probably a old default log created with docker in {filename}")
            raise e
    else:
        handler = root_logging.StreamHandler()
    handler.setFormatter(root_logging.Formatter(fmt=output_format, datefmt=datefmt))
    getLogger().addHandler(handler)
    setLevel(level)


def getEffectiveLevel():
    """Return the level the logger is set at."""
    return getLogger().getEffectiveLevel()


def flush():
    """Tell logger to empty output buffers."""
    for h in getLogger().handlers:
        h.flush()


def critical(msg, *args, **kwargs):
    """
    Log a message with severity 'CRITICAL'.
    """
    getLogger().critical(msg, *args, **kwargs)


def error(msg, *args, **kwargs):
    """
    Log a message with severity 'ERROR'.
    """
    getLogger().error(msg, *args, **kwargs)


def warning(msg, *args, **kwargs):
    """
    Log a message with severity 'WARNING'.
    """
    getLogger().warning(msg, *args, **kwargs)


def warn(msg, *args, **kwargs):
    """
    Deprecated name for warning.
    """
    getLogger().warning(msg, *args, **kwargs)


def info(msg, *args, **kwargs):
    """
    Log a message with severity 'INFO'.
    """
    getLogger().info(msg, *args, **kwargs)


def debug(msg, *args, **kwargs):
    """
    Log a message with severity 'DEBUG'.
    """
    getLogger().debug(msg, *args, **kwargs)


"""
If there is a settings file (can be empty), basicConfig will be called.
It can be customized by providing arguments as yaml dict.
"""
if COBRA_CONFIG.has_section('LOGGING') and (os.getuid() != 0):  # never automatically write logs as root
    basicConfig(**COBRA_CONFIG['LOGGING'])
else:
    basicConfig()



#!/usr/bin/env python3
# Author: Matthias Mayer
# Date: 6.7.2023
import configparser
import os
from pathlib import Path

# Get default configuration file
with open(Path(__file__).parent.joinpath('cobra.config.sample')) as f:
    DEFAULT_CONFIG_TXT = f.read()

# Find the package directory of cobra-io
this_file = Path(__file__).absolute()
parent_dir = this_file
while parent_dir.name != 'cobra-io':
    parent_dir = parent_dir.parent
    if parent_dir.name == '':
        break

# Configure configparser
COBRA_CONFIG = configparser.ConfigParser()
if parent_dir.name == 'cobra-io':  # Keep the file local if cobra-io is installed from source
    CONFIG_FILE = parent_dir.joinpath('cobra.config')
    if os.getenv("CONDA_DEFAULT_ENV") == "cobra-io":
        print("Detected conda environment cobra-io. The configuration file is not plac.")
else:
    CONFIG_FILE = Path('~/.config/cobra.config').expanduser()
    CONFIG_FILE.parent.mkdir(parents=True, exist_ok=True)
if not CONFIG_FILE.exists():
    with open(CONFIG_FILE, 'w') as f:
        f.write(DEFAULT_CONFIG_TXT)
    print("Created a default configuration file at", CONFIG_FILE, "for cobra-io.")
COBRA_CONFIG.read(CONFIG_FILE)

# Get global variables
COBRA_VERSION = COBRA_CONFIG['API']['version']



from contextlib import redirect_stdout
import io
import json
import logging
import os
from pathlib import Path
import sys
import tempfile
from unittest.mock import patch
import uuid

from timor.task import Solution, Task
import unittest

import cobra.solution
from cobra.utils import caches
from cobra.utils.requests import session


class TestSolutionAPI(unittest.TestCase):
    """Test solution API."""
    def test_solution_retrieval(self):
        with self.assertRaises(KeyError):
            cobra.solution.find_solutions(task_id='simple/PTP_1', version="some-not-existing-version")
        solution_uuids, solution_details = cobra.solution.find_solutions(task_id='simple/PTP_1')
        self.assertGreater(len(solution_uuids), 0)
        for solution_uuid, solution in zip(solution_uuids, solution_details):
            self.assertIsNotNone(solution['id'])
            self.assertIsNotNone(solution['costFunction'])
            self.assertIsNotNone(solution['cost'])
            self.assertEqual(session.get(solution['json']).status_code, 200)

            # Check details same:
            detail = cobra.solution.get_solution_details(solution_uuid)
            for k in detail.keys():
                if k in {'json', 'png', 'html'}:
                    continue  # Might change due to storage bucket creating new access tokens
                self.assertEqual(detail[k], solution[k])

    def test_solution_file_retrieval(self):
        solution_uuids, _ = cobra.solution.find_solutions(task_id='simple/PTP_1')
        first = cobra.solution.get_solution(solution_uuids[0])
        # test cache
        cached = cobra.solution.get_solution(solution_uuids[0])
        for solution_file, task_file in [first, cached]:
            self.assertTrue(solution_file.is_file())
            self.assertTrue(task_file.is_file())
            self.assertIsNotNone(json.load(solution_file.open()))
            self.assertIsNotNone(json.load(task_file.open()))
            timor_task = Task.Task.from_json_file(task_file)
            self.assertIsNotNone(timor_task)
            timor_solution = Solution.SolutionTrajectory.from_json_file(solution_file, {timor_task.id: timor_task})
            self.assertIsNotNone(timor_solution)
            self.assertIsNotNone(timor_solution.cost)

    def test_solution_retrieval_errors(self):
        with self.assertRaises(ValueError):
            cobra.solution.find_solutions()
        with self.assertRaises(KeyError):
            cobra.solution.find_solutions(task_id='non-existing-task-id')
        with self.assertRaises(ValueError):
            cobra.solution.find_solutions(task_uuid="non-existing-task-uuid")
        with self.assertRaises(ValueError):
            cobra.solution.get_solution_details(solution_uuid="non-existing-solution-uuid")

    def test_solution_upload(self,
                             test_user=os.environ.get("TEST_USER", 'cobra@cps.cit.tum.de'),
                             test_password=os.environ.get("TEST_USER_PASSWORD", '!2345678')):
        """
        Run test of solution upload / removal.

        :param test_user: User to use for testing. Is set via hidden variable in CI.
        :param test_password: Password to use for testing. Is set via hidden variable in CI.
        """
        caches.reset_cache()
        logging.info(f"Running test with user {test_user}")
        solution_uuids, solutions_data = cobra.solution.find_solutions(task_id='simple/PTP_1')
        solution_file, task_file = cobra.solution.get_solution(solution_uuids[0])
        solution_data = json.load(solution_file.open())
        # Alter such that registered as new solution
        solution_data['author'].append('CoBRA I/O Test Case ' + str(uuid.uuid1()))
        if test_user in solution_data['email']:
            solution_data['email'].remove(test_user)
        with tempfile.NamedTemporaryFile("w") as tmp_file:
            json.dump(solution_data, tmp_file)
            tmp_file.flush()
            with self.assertRaises(ValueError):  # User not in solution file
                cobra.solution.submit_solution(Path(tmp_file.name), test_user, test_password)
        with tempfile.NamedTemporaryFile("w") as tmp_file:
            solution_data['email'].append(test_user)
            json.dump(solution_data, tmp_file)
            tmp_file.flush()
            status, solution_uuid = cobra.solution.submit_solution(Path(tmp_file.name), test_user, test_password)
            self.assertTrue(status)
        retrieved_solution, _ = cobra.solution.get_solution(solution_uuid=solution_uuid)
        self.assertEqual(solution_data, json.load(retrieved_solution.open()))
        self.assertTrue(cobra.solution.delete_solution(solution_uuid, test_user, test_password))
        with self.assertRaises(ValueError):
            cobra.solution.get_solution(solution_uuid=solution_uuid, force_download=False)
        with self.assertRaises(ValueError):
            cobra.solution.get_solution(solution_uuid=solution_uuid, force_download=True)

    def test_schema_download(self):
        caches.reset_cache()
        f = cobra.solution.get_schema()
        self.assertTrue(f.is_file())
        data = json.load(f.open())
        self.assertTrue('$id' in data)
        self.assertTrue('$schema' in data)
        # Check that cached version is used
        f = cobra.solution.get_schema()
        self.assertTrue(f.is_file())
        data = json.load(f.open())
        self.assertTrue('$id' in data)
        self.assertTrue('$schema' in data)

    def test_main(self):
        stdout = io.StringIO()
        with redirect_stdout(stdout):
            with patch.object(sys, 'argv', ["solution.py", "--task_id", "simple/PTP_1"]):
                cobra.solution.main()
        self.assertTrue("solutions for task simple/PTP_1 in version" in stdout.getvalue())
        solution_uuids, _ = cobra.solution.find_solutions(task_id='simple/PTP_1')
        with redirect_stdout(stdout):
            with patch.object(sys, 'argv', ["solution.py", "--solution_uuid", solution_uuids[0]]):
                cobra.solution.main()
        self.assertTrue("Solution file: " in stdout.getvalue())
        self.assertTrue(", Task file: " in stdout.getvalue())

    def test_neg_solution(self):
        files = cobra.solution.get_negative_solutions(force_download=True)
        self.assertTrue(len(files) > 0)
        for file in files:
            self.assertTrue(file.is_file())
            self.assertTrue(file.name.endswith('.json'))
            data = json.load(file.open())
            self.assertIn('failure', data)
            self.assertIn('taskID', data)
